@@ -2,7 +2,7 @@
    +----------------------------------------------------------------------+
    | Xdebug                                                               |
    +----------------------------------------------------------------------+
-   | Copyright (c) 2002-2020 Derick Rethans                               |
+   | Copyright (c) 2002-2021 Derick Rethans                               |
    +----------------------------------------------------------------------+
    | This source file is subject to version 1.01 of the Xdebug license,   |
    | that is bundled with this package in the file LICENSE, and is        |
@@ -12,9 +12,6 @@
    | to obtain it through the world-wide-web, please send a note to       |
    | derick@xdebug.org so we can mail you a copy immediately.             |
    +----------------------------------------------------------------------+
-   | Authors: Derick Rethans <derick@xdebug.org>                          |
-   |          Shane Caraveo <shanec@ActiveState.com>                      |
-   +----------------------------------------------------------------------+
  */
 
 #include <sys/types.h>
@@ -68,6 +65,7 @@ xdebug_remote_handler xdebug_handler_dbgp = {
 	xdebug_dbgp_resolve_breakpoints,
 	xdebug_dbgp_stream_output,
 	xdebug_dbgp_notification,
+	xdebug_dbgp_user_notify,
 	xdebug_dbgp_register_eval_id,
 };
 
@@ -464,8 +462,8 @@ static int check_evaled_code(zend_string *filename_in, char **filename_out)
 	if (end_marker >= ZSTR_VAL(filename_in) && strcmp("eval()'d code", end_marker) == 0) {
 		if (xdebug_hash_find(XG_DBG(context).eval_id_lookup, ZSTR_VAL(filename_in), ZSTR_LEN(filename_in), (void *) &ei)) {
 			*filename_out = xdebug_sprintf("dbgp://%lu", ei->id);
+			return 1;
 		}
-		return 1;
 	}
 	return 0;
 }
@@ -480,7 +478,7 @@ static xdebug_xml_node* return_stackframe(int nr)
 	fse = xdebug_get_stack_frame(nr);
 	fse_prev = xdebug_get_stack_frame(nr - 1);
 
-	tmp_fname = xdebug_show_fname(fse->function, 0, 0);
+	tmp_fname = xdebug_show_fname(fse->function, 0);
 
 	tmp = xdebug_xml_node_init("stack");
 	xdebug_xml_add_attribute_ex(tmp, "where", xdstrdup(tmp_fname), 0, 1);
@@ -489,6 +487,7 @@ static xdebug_xml_node* return_stackframe(int nr)
 		if (check_evaled_code(fse_prev->filename, &tmp_filename)) {
 			xdebug_xml_add_attribute_ex(tmp, "type",     xdstrdup("eval"), 0, 1);
 			xdebug_xml_add_attribute_ex(tmp, "filename", tmp_filename, 0, 0);
+			xdfree(tmp_filename);
 		} else {
 			xdebug_xml_add_attribute_ex(tmp, "type",     xdstrdup("file"), 0, 1);
 			xdebug_xml_add_attribute_ex(tmp, "filename", xdebug_path_to_url(fse_prev->filename), 0, 1);
@@ -502,6 +501,7 @@ static xdebug_xml_node* return_stackframe(int nr)
 		if (check_evaled_code(executed_filename, &tmp_filename)) {
 			xdebug_xml_add_attribute_ex(tmp, "type", xdstrdup("eval"), 0, 1);
 			xdebug_xml_add_attribute_ex(tmp, "filename", tmp_filename, 0, 0);
+			xdfree(tmp_filename);
 		} else {
 			xdebug_xml_add_attribute_ex(tmp, "type", xdstrdup("file"), 0, 1);
 			xdebug_xml_add_attribute_ex(tmp, "filename", xdebug_path_to_url(executed_filename), 0, 1);
@@ -827,7 +827,7 @@ DBGP_FUNC(breakpoint_update)
 }
 
 
-static void breakpoint_list_helper(void *xml, xdebug_hash_element *he)
+static void breakpoint_list_helper(void *xml, xdebug_hash_element *he, void *dummy)
 {
 	xdebug_xml_node  *xml_node = (xdebug_xml_node*) xml;
 	xdebug_xml_node  *child;
@@ -842,14 +842,12 @@ static void breakpoint_list_helper(void *xml, xdebug_hash_element *he)
 
 DBGP_FUNC(breakpoint_list)
 {
-	xdebug_hash_apply(context->breakpoint_list, (void *) *retval, breakpoint_list_helper);
+	xdebug_hash_apply_with_argument(context->breakpoint_list, (void *) *retval, breakpoint_list_helper, NULL);
 }
 
 DBGP_FUNC(breakpoint_set)
 {
 	xdebug_brk_info      *brk_info;
-	char                 *tmp_name;
-	size_t                new_length = 0;
 	XDEBUG_STR_SWITCH_DECL;
 
 	brk_info = xdebug_brk_info_ctor();
@@ -888,6 +886,9 @@ DBGP_FUNC(breakpoint_set)
 	}
 
 	if ((strcmp(CMD_OPTION_CHAR('t'), "line") == 0) || (strcmp(CMD_OPTION_CHAR('t'), "conditional") == 0)) {
+		size_t  new_length = 0;
+		char   *tmp_name;
+
 		if (!CMD_OPTION_SET('n')) {
 			RETURN_RESULT(XG_DBG(status), XG_DBG(reason), XDEBUG_ERROR_INVALID_ARGS);
 		}
@@ -945,6 +946,9 @@ DBGP_FUNC(breakpoint_set)
 	} else
 
 	if ((strcmp(CMD_OPTION_CHAR('t'), "call") == 0) || (strcmp(CMD_OPTION_CHAR('t'), "return") == 0)) {
+		void *dummy = NULL;
+		char *tmp_name;
+
 		if (strcmp(CMD_OPTION_CHAR('t'), "call") == 0) {
 			brk_info->function_break_type = XDEBUG_BREAKPOINT_TYPE_CALL;
 		} else {
@@ -955,35 +959,41 @@ DBGP_FUNC(breakpoint_set)
 			RETURN_RESULT(XG_DBG(status), XG_DBG(reason), XDEBUG_ERROR_INVALID_ARGS);
 		}
 		brk_info->functionname = xdstrdup(CMD_OPTION_CHAR('m'));
-		if (CMD_OPTION_SET('a')) {
-			int   res;
 
+		if (CMD_OPTION_SET('a')) {
 			brk_info->classname = xdstrdup(CMD_OPTION_CHAR('a'));
-			tmp_name = xdebug_sprintf("%s::%s", CMD_OPTION_CHAR('a'), CMD_OPTION_CHAR('m'));
-			res = xdebug_hash_add(context->function_breakpoints, tmp_name, strlen(tmp_name), (void*) brk_info);
+			tmp_name = xdebug_sprintf(
+				"%c/%s::%s",
+				brk_info->function_break_type == XDEBUG_BREAKPOINT_TYPE_CALL ? 'C' : 'R',
+				CMD_OPTION_CHAR('a'), CMD_OPTION_CHAR('m')
+			);
+		} else {
+			tmp_name = xdebug_sprintf(
+				"%c/%s",
+				brk_info->function_break_type == XDEBUG_BREAKPOINT_TYPE_CALL ? 'C' : 'R',
+				CMD_OPTION_CHAR('m')
+			);
+		}
+
+		if (xdebug_hash_find(context->function_breakpoints, tmp_name, strlen(tmp_name), (void*) &dummy)) {
+			xdfree(tmp_name);
+			RETURN_RESULT(XG_DBG(status), XG_DBG(reason), XDEBUG_ERROR_BREAKPOINT_NOT_SET);
+		}
+
+		if (!xdebug_hash_add(context->function_breakpoints, tmp_name, strlen(tmp_name), brk_info)) {
+			xdfree(tmp_name);
+			RETURN_RESULT(XG_DBG(status), XG_DBG(reason), XDEBUG_ERROR_BREAKPOINT_NOT_SET);
+		} else {
 			if (brk_info->function_break_type == XDEBUG_BREAKPOINT_TYPE_CALL) {
 				brk_info->id = breakpoint_admin_add(context, XDEBUG_BREAKPOINT_TYPE_CALL, tmp_name);
 			} else {
 				brk_info->id = breakpoint_admin_add(context, XDEBUG_BREAKPOINT_TYPE_RETURN, tmp_name);
 			}
-			xdfree(tmp_name);
-
-			if (!res) {
-				RETURN_RESULT(XG_DBG(status), XG_DBG(reason), XDEBUG_ERROR_INVALID_ARGS);
-			}
-		} else {
-			if (!xdebug_hash_add(context->function_breakpoints, CMD_OPTION_CHAR('m'), CMD_OPTION_LEN('m'), (void*) brk_info)) {
-				RETURN_RESULT(XG_DBG(status), XG_DBG(reason), XDEBUG_ERROR_BREAKPOINT_NOT_SET);
-			} else {
-				if (brk_info->function_break_type == XDEBUG_BREAKPOINT_TYPE_CALL) {
-					brk_info->id = breakpoint_admin_add(context, XDEBUG_BREAKPOINT_TYPE_CALL, CMD_OPTION_CHAR('m'));
-				} else {
-					brk_info->id = breakpoint_admin_add(context, XDEBUG_BREAKPOINT_TYPE_RETURN, CMD_OPTION_CHAR('m'));
-				}
-			}
 		}
 
 		brk_info->resolved = XDEBUG_BRK_RESOLVED;
+
+		xdfree(tmp_name);
 	} else
 
 	if (strcmp(CMD_OPTION_CHAR('t'), "exception") == 0) {
@@ -1037,7 +1047,7 @@ DBGP_FUNC(eval)
 	xdfree(eval_string);
 
 	/* Handle result */
-	if (res == FAILURE) {
+	if (!res) {
 		RETURN_RESULT(XG_DBG(status), XG_DBG(reason), XDEBUG_ERROR_EVALUATING_CODE);
 	} else {
 		ret_xml = xdebug_get_zval_value_xml_node(NULL, &ret_zval, options);
@@ -1151,6 +1161,11 @@ DBGP_FUNC(detach)
 	xdebug_mark_debug_connection_not_active();
 	XG_DBG(stdout_mode) = 0;
 	XG_DBG(detached) = 1;
+
+	if (CMD_OPTION_SET('-')) {
+		XG_DBG(context).detached_message = xdstrdup(CMD_OPTION_CHAR('-'));
+		xdebug_log_ex(XLOG_CHAN_DEBUG, XLOG_WARN, "DETACH", "Debug client detached: %s.", XG_DBG(context).detached_message);
+	}
 }
 
 
@@ -1235,7 +1250,7 @@ DBGP_FUNC(feature_get)
 		XDEBUG_STR_CASE_END
 
 		XDEBUG_STR_CASE("language_version")
-			xdebug_xml_add_text(*retval, xdstrdup(PHP_VERSION));
+			xdebug_xml_add_text(*retval, xdstrdup(XG_BASE(php_version_run_time)));
 			xdebug_xml_add_attribute(*retval, "supported", "1");
 		XDEBUG_STR_CASE_END
 
@@ -1294,6 +1309,11 @@ DBGP_FUNC(feature_get)
 			xdebug_xml_add_attribute(*retval, "supported", "1");
 		XDEBUG_STR_CASE_END
 
+		XDEBUG_STR_CASE("breakpoint_details")
+			xdebug_xml_add_text(*retval, xdebug_sprintf("%ld", XG_DBG(context).breakpoint_details));
+			xdebug_xml_add_attribute(*retval, "supported", "1");
+		XDEBUG_STR_CASE_END
+
 		XDEBUG_STR_CASE_DEFAULT
 			xdebug_xml_add_text(*retval, xdstrdup(lookup_cmd(CMD_OPTION_CHAR('n')) ? "1" : "0"));
 			xdebug_xml_add_attribute(*retval, "supported", lookup_cmd(CMD_OPTION_CHAR('n')) ? "1" : "0");
@@ -1361,6 +1381,10 @@ DBGP_FUNC(feature_set)
 			XG_DBG(context).resolved_breakpoints = strtol(CMD_OPTION_CHAR('v'), NULL, 10);
 		XDEBUG_STR_CASE_END
 
+		XDEBUG_STR_CASE("breakpoint_details")
+			XG_DBG(context).breakpoint_details = strtol(CMD_OPTION_CHAR('v'), NULL, 10);
+		XDEBUG_STR_CASE_END
+
 		XDEBUG_STR_CASE_DEFAULT
 			RETURN_RESULT(XG_DBG(status), XG_DBG(reason), XDEBUG_ERROR_INVALID_ARGS);
 		XDEBUG_STR_CASE_DEFAULT_END
@@ -1455,7 +1479,13 @@ DBGP_FUNC(property_get)
 	/* Set the symbol table corresponding with the requested stack depth */
 	if (context_nr == 0) { /* locals */
 		if ((fse = xdebug_get_stack_frame(depth))) {
-			function_stack_entry *old_fse = xdebug_get_stack_frame(depth - 1);
+			function_stack_entry *old_fse;
+
+			if (!fse->user_defined) {
+				RETURN_RESULT(XG_DBG(status), XG_DBG(reason), XDEBUG_ERROR_PROPERTY_NON_EXISTENT);
+			}
+
+			old_fse = xdebug_get_stack_frame(depth - 1);
 
 			if (depth > 0) {
 				xdebug_lib_set_active_data(old_fse->execute_data);
@@ -1619,7 +1649,7 @@ DBGP_FUNC(property_set)
 	xdfree(new_value);
 
 	/* Handle result */
-	if (res == FAILURE) {
+	if (!res) {
 		/* don't send an error, send success = zero */
 		xdebug_xml_add_attribute(*retval, "success", "0");
 	} else {
@@ -1846,7 +1876,7 @@ static int attach_context_vars(xdebug_xml_node *node, xdebug_var_export_options
 		 * method call as we attach constants and static properties to "this"
 		 * too normally. */
 		if (fse->function.type == XFUNC_STATIC_MEMBER) {
-			zend_class_entry *ce = zend_fetch_class(fse->function.class_name, ZEND_FETCH_CLASS_DEFAULT);
+			zend_class_entry *ce = zend_fetch_class(fse->function.object_class, ZEND_FETCH_CLASS_DEFAULT);
 
 #if PHP_VERSION_ID >= 70400
 			if (ce->type == ZEND_INTERNAL_CLASS || (ce->ce_flags & ZEND_ACC_IMMUTABLE)) {
@@ -1867,7 +1897,7 @@ static int attach_context_vars(xdebug_xml_node *node, xdebug_var_export_options
 
 DBGP_FUNC(stack_depth)
 {
-	xdebug_xml_add_attribute_ex(*retval, "depth", xdebug_sprintf("%lu", XG_BASE(level)), 0, 1);
+	xdebug_xml_add_attribute_ex(*retval, "depth", xdebug_sprintf("%lu", XDEBUG_VECTOR_COUNT(XG_BASE(stack))), 0, 1);
 }
 
 DBGP_FUNC(stack_get)
@@ -1877,7 +1907,7 @@ DBGP_FUNC(stack_get)
 
 	if (CMD_OPTION_SET('d')) {
 		depth = strtol(CMD_OPTION_CHAR('d'), NULL, 10);
-		if (depth >= 0 && depth < (long) XG_BASE(level)) {
+		if (depth >= 0 && depth < (long) XDEBUG_VECTOR_COUNT(XG_BASE(stack))) {
 			stackframe = return_stackframe(depth);
 			xdebug_xml_add_child(*retval, stackframe);
 		} else {
@@ -1970,7 +2000,7 @@ DBGP_FUNC(xcmd_get_executable_lines)
 	}
 
 	depth = strtol(CMD_OPTION_CHAR('d'), NULL, 10);
-	if (depth >= 0 && depth < (long) XG_BASE(level)) {
+	if (depth >= 0 && depth < (long) XDEBUG_VECTOR_COUNT(XG_BASE(stack))) {
 		fse = xdebug_get_stack_frame(depth);
 	} else {
 		RETURN_RESULT(XG_DBG(status), XG_DBG(reason), XDEBUG_ERROR_STACK_DEPTH_INVALID);
@@ -2323,6 +2353,14 @@ static int xdebug_dbgp_cmdloop(xdebug_con *context, int bail)
 
 }
 
+static int xdebug_compare_brk_info(const void *le1, const void *le2)
+{
+	xdebug_brk_info *a = (xdebug_brk_info *) XDEBUG_LLIST_VALP(*(xdebug_llist_element **) le1);
+	xdebug_brk_info *b = (xdebug_brk_info *) XDEBUG_LLIST_VALP(*(xdebug_llist_element **) le2);
+
+	return (a->id > b->id);
+}
+
 int xdebug_dbgp_init(xdebug_con *context, int mode)
 {
 	xdebug_var_export_options *options;
@@ -2368,7 +2406,7 @@ int xdebug_dbgp_init(xdebug_con *context, int mode)
 		xdebug_xml_add_attribute_ex(response, "fileuri", xdebug_path_to_url(context->program_name), 0, 1);
 	}
 	xdebug_xml_add_attribute_ex(response, "language", "PHP", 0, 0);
-	xdebug_xml_add_attribute_ex(response, "xdebug:language_version", PHP_VERSION, 0, 0);
+	xdebug_xml_add_attribute_ex(response, "xdebug:language_version", XG_BASE(php_version_run_time), 0, 0);
 	xdebug_xml_add_attribute_ex(response, "protocol_version", DBGP_VERSION, 0, 0);
 	xdebug_xml_add_attribute_ex(response, "appid", xdebug_sprintf(ZEND_ULONG_FMT, xdebug_get_pid()), 0, 1);
 
@@ -2376,10 +2414,13 @@ int xdebug_dbgp_init(xdebug_con *context, int mode)
 		xdebug_xml_add_attribute_ex(response, "session", xdstrdup(getenv("DBGP_COOKIE")), 0, 1);
 	}
 
-	if (XINI_DBG(cloud_id) && *XINI_DBG(cloud_id)) {
+	if (XG_DBG(context).host_type == XDEBUG_CLOUD && XINI_DBG(cloud_id) && *XINI_DBG(cloud_id)) {
 		xdebug_xml_add_attribute_ex(response, "xdebug:userid", xdstrdup(XINI_DBG(cloud_id)), 0, 1);
 	}
-	if (XG_DBG(ide_key) && *XG_DBG(ide_key)) {
+	if (XG_DBG(context).host_type == XDEBUG_CLOUD_FROM_TRIGGER_VALUE && XG_DBG(ide_key) && *XG_DBG(ide_key)) {
+		xdebug_xml_add_attribute_ex(response, "xdebug:userid", xdstrdup(XG_DBG(ide_key)), 0, 1);
+	}
+	if (XG_DBG(context).host_type == XDEBUG_NORMAL && XG_DBG(ide_key) && *XG_DBG(ide_key)) {
 		xdebug_xml_add_attribute_ex(response, "idekey", xdstrdup(XG_DBG(ide_key)), 0, 1);
 	}
 
@@ -2405,7 +2446,7 @@ int xdebug_dbgp_init(xdebug_con *context, int mode)
 		options->runtime[i].current_element_nr = 0;
 	}
 
-	context->breakpoint_list = xdebug_hash_alloc(64, (xdebug_hash_dtor_t) xdebug_hash_admin_dtor);
+	context->breakpoint_list = xdebug_hash_alloc_with_sort(64, (xdebug_hash_dtor_t) xdebug_hash_admin_dtor, xdebug_compare_brk_info);
 	context->function_breakpoints = xdebug_hash_alloc(64, (xdebug_hash_dtor_t) xdebug_hash_brk_dtor);
 	context->exception_breakpoints = xdebug_hash_alloc(64, (xdebug_hash_dtor_t) xdebug_hash_brk_dtor);
 	context->line_breakpoints = xdebug_llist_alloc((xdebug_llist_dtor) xdebug_llist_brk_dtor);
@@ -2414,6 +2455,7 @@ int xdebug_dbgp_init(xdebug_con *context, int mode)
 	context->send_notifications = 0;
 	context->inhibit_notifications = 0;
 	context->resolved_breakpoints = 0;
+	context->breakpoint_details = 0;
 
 	xdebug_mark_debug_connection_active();
 	xdebug_dbgp_cmdloop(context, XDEBUG_CMDLOOP_BAIL);
@@ -2541,8 +2583,9 @@ int xdebug_dbgp_error(xdebug_con *context, int type, char *exception_type, char
 
 int xdebug_dbgp_break_on_line(xdebug_con *context, xdebug_brk_info *brk, zend_string *filename, int lineno)
 {
-	char *tmp_file     = ZSTR_VAL(filename);
-	int   tmp_file_len = ZSTR_LEN(filename);
+	char *tmp_file      = ZSTR_VAL(filename);
+	int   tmp_file_len  = ZSTR_LEN(filename);
+	int   free_tmp_file = 0;
 
 	xdebug_log(XLOG_CHAN_DEBUG, XLOG_DEBUG, "Checking whether to break on %s:%d.", ZSTR_VAL(brk->filename), brk->resolved_lineno);
 
@@ -2554,6 +2597,7 @@ int xdebug_dbgp_break_on_line(xdebug_con *context, xdebug_brk_info *brk, zend_st
 	xdebug_log(XLOG_CHAN_DEBUG, XLOG_DEBUG, "I: Current location: %s:%d.", tmp_file, lineno);
 
 	if (is_dbgp_url(brk->filename) && check_evaled_code(filename, &tmp_file)) {
+		free_tmp_file = 1;
 		tmp_file_len = strlen(tmp_file);
 		xdebug_log(XLOG_CHAN_DEBUG, XLOG_DEBUG, "I: Found eval code for '%s': %s.", ZSTR_VAL(filename), tmp_file);
 	}
@@ -2562,24 +2606,40 @@ int xdebug_dbgp_break_on_line(xdebug_con *context, xdebug_brk_info *brk, zend_st
 
 	if (ZSTR_LEN(brk->filename) != tmp_file_len) {
 		xdebug_log(XLOG_CHAN_DEBUG, XLOG_DEBUG, "R: File name length (%d) doesn't match with breakpoint (%zd).", tmp_file_len, ZSTR_LEN(brk->filename));
+
+		if (free_tmp_file) {
+			xdfree(tmp_file);
+		}
 		return 0;
 	}
 
 	if (brk->resolved_lineno != lineno) {
 		xdebug_log(XLOG_CHAN_DEBUG, XLOG_DEBUG, "R: Line number (%d) doesn't match with breakpoint (%d).", lineno, brk->resolved_lineno);
+
+		if (free_tmp_file) {
+			xdfree(tmp_file);
+		}
 		return 0;
 	}
 
 	if (strncasecmp(ZSTR_VAL(brk->filename), tmp_file, ZSTR_LEN(brk->filename)) == 0) {
 		xdebug_log(XLOG_CHAN_DEBUG, XLOG_DEBUG, "F: File names match (%s).", ZSTR_VAL(brk->filename));
+
+		if (free_tmp_file) {
+			xdfree(tmp_file);
+		}
 		return 1;
 	}
 
 	xdebug_log(XLOG_CHAN_DEBUG, XLOG_DEBUG, "R: File names (%s) doesn't match with breakpoint (%s).", tmp_file, ZSTR_VAL(brk->filename));
+
+	if (free_tmp_file) {
+		xdfree(tmp_file);
+	}
 	return 0;
 }
 
-int xdebug_dbgp_breakpoint(xdebug_con *context, xdebug_vector *stack, zend_string *filename, long lineno, int type, char *exception, char *code, char *message)
+int xdebug_dbgp_breakpoint(xdebug_con *context, xdebug_vector *stack, zend_string *filename, long lineno, int type, char *exception, char *code, const char *message, xdebug_brk_info *brk_info)
 {
 	xdebug_xml_node *response, *error_container;
 
@@ -2604,6 +2664,7 @@ int xdebug_dbgp_breakpoint(xdebug_con *context, xdebug_vector *stack, zend_strin
 
 		if (check_evaled_code(filename, &tmp_filename)) {
 			xdebug_xml_add_attribute_ex(error_container, "filename", tmp_filename, 0, 0);
+			xdfree(tmp_filename);
 		} else {
 			xdebug_xml_add_attribute_ex(error_container, "filename", xdebug_path_to_url(filename), 0, 1);
 		}
@@ -2622,6 +2683,12 @@ int xdebug_dbgp_breakpoint(xdebug_con *context, xdebug_vector *stack, zend_strin
 	}
 	xdebug_xml_add_child(response, error_container);
 
+	if (XG_DBG(context).breakpoint_details && brk_info) {
+		xdebug_xml_node *breakpoint_node = xdebug_xml_node_init("breakpoint");
+		breakpoint_brk_info_add(breakpoint_node, brk_info);
+		xdebug_xml_add_child(response, breakpoint_node);
+	}
+
 	send_message(context, response);
 	xdebug_xml_node_dtor(response);
 
@@ -2783,7 +2850,7 @@ static void line_breakpoint_resolve_helper(xdebug_con *context, xdebug_lines_lis
 	}
 }
 
-static void breakpoint_resolve_helper(void *rctxt, xdebug_hash_element *he)
+static void breakpoint_resolve_helper(void *rctxt, xdebug_hash_element *he, void *dummy)
 {
 	xdebug_dbgp_resolve_context *ctxt = (xdebug_dbgp_resolve_context*) rctxt;
 	xdebug_brk_admin            *admin = (xdebug_brk_admin*) he->ptr;
@@ -2837,7 +2904,7 @@ int xdebug_dbgp_resolve_breakpoints(xdebug_con *context, zend_string *filename)
 	resolv_ctxt.context = context;
 	resolv_ctxt.filename = filename;
 	resolv_ctxt.lines_list = lines_list;
-	xdebug_hash_apply(context->breakpoint_list, (void *) &resolv_ctxt, breakpoint_resolve_helper);
+	xdebug_hash_apply_with_argument(context->breakpoint_list, (void *) &resolv_ctxt, breakpoint_resolve_helper, NULL);
 
 	return 1;
 }
@@ -2869,6 +2936,7 @@ int xdebug_dbgp_notification(xdebug_con *context, zend_string *filename, long li
 
 		if (check_evaled_code(filename, &tmp_filename)) {
 			xdebug_xml_add_attribute_ex(error_container, "filename", tmp_filename, 0, 0);
+			xdfree(tmp_filename);
 		} else {
 			xdebug_xml_add_attribute_ex(error_container, "filename", xdebug_path_to_url(filename), 0, 1);
 		}
@@ -2896,6 +2964,49 @@ int xdebug_dbgp_notification(xdebug_con *context, zend_string *filename, long li
 	return 1;
 }
 
+int xdebug_dbgp_user_notify(xdebug_con *context, zend_string *filename, long lineno, zval *data)
+{
+	xdebug_xml_node *response, *data_node, *location_node;
+	xdebug_var_export_options *options;
+
+	if (!context->send_notifications) {
+		return 0;
+	}
+
+	response = xdebug_xml_node_init("notify");
+	xdebug_xml_add_attribute(response, "xmlns", "urn:debugger_protocol_v1");
+	xdebug_xml_add_attribute(response, "xmlns:xdebug", "https://xdebug.org/dbgp/xdebug");
+	xdebug_xml_add_attribute(response, "name", "user");
+
+	options = (xdebug_var_export_options*) context->options;
+	options->encode_as_extended_property = 0;
+
+	location_node = xdebug_xml_node_init("xdebug:location");
+	if (filename) {
+		char *tmp_filename = NULL;
+
+		if (check_evaled_code(filename, &tmp_filename)) {
+			xdebug_xml_add_attribute_ex(location_node, "filename", tmp_filename, 0, 0);
+			xdfree(tmp_filename);
+		} else {
+			xdebug_xml_add_attribute_ex(location_node, "filename", xdebug_path_to_url(filename), 0, 1);
+		}
+	}
+	if (lineno) {
+		xdebug_xml_add_attribute_ex(location_node, "lineno", xdebug_sprintf("%lu", lineno), 0, 1);
+	}
+	xdebug_xml_add_child(response, location_node);
+
+	data_node = xdebug_xml_node_init("property");
+	xdebug_var_export_xml_node(&data, NULL, data_node, options, 0);
+	xdebug_xml_add_child(response, data_node);
+
+	send_message(context, response);
+	xdebug_xml_node_dtor(response);
+
+	return 1;
+}
+
 static char *create_eval_key_file(zend_string *filename, int lineno)
 {
 	return xdebug_sprintf("%s(%d) : eval()'d code", ZSTR_VAL(filename), lineno);
@@ -2920,9 +3031,11 @@ int xdebug_dbgp_register_eval_id(xdebug_con *context, function_stack_entry *fse)
 
 	key = create_eval_key_file(fse->filename, fse->lineno);
 	xdebug_hash_add(context->eval_id_lookup, key, strlen(key), (void*) ei);
+	xdfree(key);
 
 	key = create_eval_key_id(ei->id);
 	xdebug_hash_add(context->eval_id_lookup, key, strlen(key), (void*) ei);
+	xdfree(key);
 
 	return ei->id;
 }
