@@ -2,7 +2,7 @@
    +----------------------------------------------------------------------+
    | Xdebug                                                               |
    +----------------------------------------------------------------------+
-   | Copyright (c) 2002-2020 Derick Rethans                               |
+   | Copyright (c) 2002-2022 Derick Rethans                               |
    +----------------------------------------------------------------------+
    | This source file is subject to version 1.01 of the Xdebug license,   |
    | that is bundled with this package in the file LICENSE, and is        |
@@ -12,8 +12,6 @@
    | to obtain it through the world-wide-web, please send a note to       |
    | derick@xdebug.org so we can mail you a copy immediately.             |
    +----------------------------------------------------------------------+
-   | Authors: Derick Rethans <derick@xdebug.org>                          |
-   +----------------------------------------------------------------------+
  */
 
 #include "php_xdebug.h"
@@ -27,8 +25,9 @@ extern ZEND_DECLARE_MODULE_GLOBALS(xdebug);
 
 void xdebug_init_library_globals(xdebug_library_globals_t *xg)
 {
-	xg->headers              = NULL;
-	xg->mode                 = 0xFFFFFFFF;
+	xg->headers               = NULL;
+	xg->mode                  = 0xFFFFFFFF;
+	xg->mode_from_environment = 0;
 
 	xg->log_file             = 0;
 
@@ -56,6 +55,11 @@ void xdebug_library_zend_shutdown(void)
 void xdebug_library_minit(void)
 {
 	xdebug_set_opcode_multi_handler(ZEND_ASSIGN);
+	xdebug_set_opcode_multi_handler(ZEND_ASSIGN_DIM);
+	xdebug_set_opcode_multi_handler(ZEND_ASSIGN_OBJ);
+#if PHP_VERSION_ID >= 70400
+	xdebug_set_opcode_multi_handler(ZEND_ASSIGN_STATIC_PROP);
+#endif
 	xdebug_set_opcode_multi_handler(ZEND_QM_ASSIGN);
 	xdebug_set_opcode_multi_handler(ZEND_INCLUDE_OR_EVAL);
 }
@@ -92,7 +96,6 @@ void xdebug_library_rinit(void)
 
 	XG_LIB(dumped) = 0;
 	XG_LIB(do_collect_errors) = 0;
-
 }
 
 void xdebug_library_post_deactivate(void)
@@ -180,23 +183,33 @@ int xdebug_lib_set_mode(char *mode)
 	char *config = getenv("XDEBUG_MODE");
 	int   result = 0;
 
+	/* XDEBUG_MODE environment variable */
 	if (config && strlen(config)) {
 		result = xdebug_lib_set_mode_from_setting(config);
 
 		if (!result) {
-			xdebug_log_ex(XLOG_CHAN_CONFIG, XLOG_CRIT, "ENVMODE", "Invalid mode '%s' set for 'XDEBUG_MODE' environment variable", config);
-		}
-	} else {
-		result = xdebug_lib_set_mode_from_setting(mode);
-
-		if (!result) {
-			xdebug_log_ex(XLOG_CHAN_CONFIG, XLOG_CRIT, "MODE", "Invalid mode '%s' set for 'xdebug.mode' configuration setting", mode);
+			xdebug_log_ex(XLOG_CHAN_CONFIG, XLOG_CRIT, "ENVMODE", "Invalid mode '%s' set for 'XDEBUG_MODE' environment variable, fall back to 'xdebug.mode' configuration setting", config);
+		} else {
+			XG_LIB(mode_from_environment) = 1;
+			return result;
 		}
 	}
 
+	/* 'xdebug.mode' configuration setting */
+	result = xdebug_lib_set_mode_from_setting(mode);
+
+	if (!result) {
+		xdebug_log_ex(XLOG_CHAN_CONFIG, XLOG_CRIT, "MODE", "Invalid mode '%s' set for 'xdebug.mode' configuration setting", mode);
+	}
+
 	return result;
 }
 
+int xdebug_lib_get_start_with_request(void)
+{
+	return XG_LIB(start_with_request);
+}
+
 int xdebug_lib_set_start_with_request(char *value)
 {
 	if (strcmp(value, "default") == 0) {
@@ -219,14 +232,14 @@ int xdebug_lib_set_start_with_request(char *value)
 	return 0;
 }
 
-int xdebug_lib_start_with_request(void)
+int xdebug_lib_start_with_request(int for_mode)
 {
 	if (XG_LIB(start_with_request) == XDEBUG_START_WITH_REQUEST_YES) {
 		return 1;
 	}
 
 	if (XG_LIB(start_with_request) == XDEBUG_START_WITH_REQUEST_DEFAULT) {
-		if (XDEBUG_MODE_IS(XDEBUG_MODE_PROFILING)) {
+		if (for_mode == XDEBUG_MODE_PROFILING && XDEBUG_MODE_IS(XDEBUG_MODE_PROFILING)) {
 			return 1;
 		}
 	}
@@ -244,6 +257,11 @@ int xdebug_lib_never_start_with_request(void)
 }
 
 
+int xdebug_lib_get_start_upon_error(void)
+{
+	return XG_LIB(start_upon_error);
+}
+
 int xdebug_lib_set_start_upon_error(char *value)
 {
 	if (strcmp(value, "default") == 0) {
@@ -271,10 +289,34 @@ int xdebug_lib_start_upon_error(void)
 	return 0;
 }
 
+const char *xdebug_lib_mode_from_value(int mode)
+{
+	switch (mode) {
+		case XDEBUG_MODE_DEVELOP:
+			return "develop";
+		case XDEBUG_MODE_COVERAGE:
+			return "coverage";
+		case XDEBUG_MODE_STEP_DEBUG:
+			return "debug";
+		case XDEBUG_MODE_GCSTATS:
+			return "gcstats";
+		case XDEBUG_MODE_PROFILING:
+			return "profile";
+		case XDEBUG_MODE_TRACING:
+			return "trace";
+		default:
+			return "?";
+	}
+}
 
-static zval *find_in_globals(const char *element)
+static const char *find_in_globals(const char *element)
 {
 	zval *trigger_val = NULL;
+	const char *env_value = getenv(element);
+
+	if (env_value) {
+		return env_value;
+	}
 
 	if (
 		(
@@ -287,69 +329,145 @@ static zval *find_in_globals(const char *element)
 			(trigger_val = zend_hash_str_find(Z_ARR(PG(http_globals)[TRACK_VARS_COOKIE]), element, strlen(element))) != NULL
 		)
 	) {
-		return trigger_val;
+		return Z_STRVAL_P(trigger_val);
 	}
 
 	return NULL;
 }
 
-static int trigger_enabled(char **found_trigger_value)
+int xdebug_lib_has_shared_secret(void)
 {
 	char *shared_secret = XINI_LIB(trigger_value);
-	zval *z_found_trigger_value = NULL;
 
-	/* First we check for the generic 'XDEBUG_TRIGGER' option */
-	z_found_trigger_value = find_in_globals("XDEBUG_TRIGGER");
-
-	/* If not found, we fall back to the per-mode name for backwards compatibility reasons */
-	if (!z_found_trigger_value) {
-		const char *fallback_name = NULL;
-
-		if (XG_LIB(mode) & XDEBUG_MODE_PROFILING) {
-			fallback_name = "XDEBUG_PROFILE";
-		} else if (XG_LIB(mode) & XDEBUG_MODE_TRACING) {
-			fallback_name = "XDEBUG_TRACE";
-		} else if (XG_LIB(mode) & XDEBUG_MODE_STEP_DEBUG) {
-			fallback_name = "XDEBUG_SESSION";
-		}
-
-		if (fallback_name) {
-			z_found_trigger_value = find_in_globals(fallback_name);
-		}
-	}
-
-	if (!z_found_trigger_value) {
-		if (found_trigger_value != NULL) {
-			*found_trigger_value = NULL;
-		}
-		return 0;
-	}
-
-	/* If the configured shared secret trigger value is empty, then it always
-	 * triggers */
-	if (shared_secret == NULL || shared_secret[0] == '\0') {
-		if (found_trigger_value != NULL) {
-			*found_trigger_value = xdstrdup(Z_STRVAL_P(z_found_trigger_value));
-		}
-		return 1;
-	}
-
-	/* Check if the configured trigger value matches the one found in the
-	 * trigger element */
-	if (strcmp(shared_secret, Z_STRVAL_P(z_found_trigger_value)) == 0) {
-		if (found_trigger_value != NULL) {
-			*found_trigger_value = xdstrdup(Z_STRVAL_P(z_found_trigger_value));
-		}
+	if (shared_secret != NULL && shared_secret[0] != '\0') {
 		return 1;
 	}
 
 	return 0;
 }
 
-static int is_mode_trigger_and_enabled(int force_trigger, char **found_trigger_value)
+static int does_shared_secret_match_single(int mode, const char *trimmed_trigger_value, const char *trimmed_shared_secret, char **found_trigger_value)
+{
+	if (strcmp(trimmed_shared_secret, trimmed_trigger_value) == 0) {
+		xdebug_log_ex(XLOG_CHAN_CONFIG, XLOG_DEBUG, "TRGSEC-MATCH", "The trigger value '%s' matched the shared secret '%s' for mode '%s'", trimmed_trigger_value, trimmed_shared_secret, xdebug_lib_mode_from_value(mode));
+
+		if (found_trigger_value != NULL) {
+			*found_trigger_value = xdstrdup(trimmed_trigger_value);
+		}
+
+		return 1;
+	}
+
+	return 0;
+}
+
+static int does_shared_secret_match(int mode, const char *trigger_name, const char *trigger_value, char **found_trigger_value)
+{
+	int         retval = 0;
+	const char *shared_secret = XINI_LIB(trigger_value);
+	char       *trimmed_trigger_value = xdebug_trim(trigger_value);
+
+	/* Check we have a multi-value-shared secret setting */
+	if (strchr(shared_secret, ',') != NULL) {
+		int         i;
+		xdebug_arg *values = xdebug_arg_ctor();
+
+		xdebug_log_ex(XLOG_CHAN_CONFIG, XLOG_DEBUG, "TRGSEC-MULT", "The shared secret (xdebug.trigger_value) is multi-value for mode '%s'", xdebug_lib_mode_from_value(mode));
+
+		xdebug_explode(",", shared_secret, values, -1);
+
+		for (i = 0; i < values->c; i++) {
+			char *trimmed_shared_secret = xdebug_trim(values->args[i]);
+
+			retval = does_shared_secret_match_single(mode, trimmed_trigger_value, trimmed_shared_secret, found_trigger_value);
+
+			xdfree(trimmed_shared_secret);
+
+			/* Jump out of the loop if we found a match */
+			if (retval != 0) {
+				break;
+			}
+		}
+
+		xdebug_arg_dtor(values);
+
+		if (retval == 0) {
+			xdebug_log_ex(XLOG_CHAN_CONFIG, XLOG_WARN, "TRGSEC-MNO", "The trigger value '%s', as set through '%s', did not match any of the shared secrets (xdebug.trigger_value) for mode '%s'", trimmed_trigger_value, trigger_name, xdebug_lib_mode_from_value(mode));
+		}
+	} else {
+		char *trimmed_shared_secret = xdebug_trim(shared_secret);
+
+		retval = does_shared_secret_match_single(mode, trimmed_trigger_value, trimmed_shared_secret, found_trigger_value);
+
+		xdfree(trimmed_shared_secret);
+
+		if (retval == 0) {
+			xdebug_log_ex(XLOG_CHAN_CONFIG, XLOG_WARN, "TRGSEC-NO", "The trigger value '%s', as set through '%s', did not match the shared secret (xdebug.trigger_value) for mode '%s'", trimmed_trigger_value, trigger_name, xdebug_lib_mode_from_value(mode));
+		}
+	}
+
+	xdfree(trimmed_trigger_value);
+
+	return retval;
+}
+
+static int trigger_enabled(int for_mode, char **found_trigger_value)
+{
+	const char *trigger_value = NULL;
+	const char *trigger_name = "XDEBUG_TRIGGER";
+
+	xdebug_log(XLOG_CHAN_CONFIG, XLOG_DEBUG, "Checking if trigger 'XDEBUG_TRIGGER' is enabled for mode '%s'", xdebug_lib_mode_from_value(for_mode));
+
+	/* First we check for the generic 'XDEBUG_TRIGGER' option */
+	trigger_value = find_in_globals(trigger_name);
+
+	/* If not found, we fall back to the per-mode name for backwards compatibility reasons */
+	if (!trigger_value) {
+		if (XDEBUG_MODE_IS(XDEBUG_MODE_PROFILING) && (for_mode == XDEBUG_MODE_PROFILING)) {
+			trigger_name = "XDEBUG_PROFILE";
+		} else if (XDEBUG_MODE_IS(XDEBUG_MODE_TRACING) && (for_mode == XDEBUG_MODE_TRACING)) {
+			trigger_name = "XDEBUG_TRACE";
+		} else if (XDEBUG_MODE_IS(XDEBUG_MODE_STEP_DEBUG) && (for_mode == XDEBUG_MODE_STEP_DEBUG)) {
+			trigger_name = "XDEBUG_SESSION";
+		}
+
+		if (trigger_name) {
+			xdebug_log(XLOG_CHAN_CONFIG, XLOG_INFO, "Trigger value for 'XDEBUG_TRIGGER' not found, falling back to '%s'", trigger_name);
+			trigger_value = find_in_globals(trigger_name);
+		}
+	}
+
+	if (!trigger_value) {
+		xdebug_log(XLOG_CHAN_CONFIG, XLOG_INFO, "Trigger value for '%s' not found, so not activating", trigger_name);
+
+		if (found_trigger_value != NULL) {
+			*found_trigger_value = NULL;
+		}
+		return 0;
+	}
+
+	/* If there is no configured shared secret trigger, always trigger */
+	if (!xdebug_lib_has_shared_secret()) {
+		xdebug_log(XLOG_CHAN_CONFIG, XLOG_INFO, "No shared secret: Activating");
+		if (found_trigger_value != NULL) {
+			*found_trigger_value = xdstrdup(trigger_value);
+		}
+		return 1;
+	}
+
+	/* Check if the configured trigger value matches the one found in the
+	 * trigger element */
+	if (does_shared_secret_match(for_mode, trigger_name, trigger_value, found_trigger_value)) {
+		return 1;
+	}
+
+	return 0;
+}
+
+static int is_mode_trigger_and_enabled(int for_mode, int force_trigger, char **found_trigger_value)
 {
 	if (XG_LIB(start_with_request) == XDEBUG_START_WITH_REQUEST_TRIGGER) {
-		return force_trigger || trigger_enabled(found_trigger_value);
+		return force_trigger || trigger_enabled(for_mode, found_trigger_value);
 	}
 
 	if (XG_LIB(start_with_request) == XDEBUG_START_WITH_REQUEST_DEFAULT) {
@@ -357,7 +475,7 @@ static int is_mode_trigger_and_enabled(int force_trigger, char **found_trigger_v
 			XDEBUG_MODE_IS(XDEBUG_MODE_STEP_DEBUG) ||
 			XDEBUG_MODE_IS(XDEBUG_MODE_TRACING)
 		) {
-			return force_trigger || trigger_enabled(found_trigger_value);
+			return force_trigger || trigger_enabled(for_mode, found_trigger_value);
 		}
 	}
 
@@ -367,16 +485,16 @@ static int is_mode_trigger_and_enabled(int force_trigger, char **found_trigger_v
 /* Returns 1 if the mode is 'trigger', or 'default', where the default mode for
  * a feature is to trigger, and the trigger is present. If found_trigger_value
  * is not NULL, then it is set to the found trigger value */
-int xdebug_lib_start_with_trigger(char **found_trigger_value)
+int xdebug_lib_start_with_trigger(int for_mode, char **found_trigger_value)
 {
-	return is_mode_trigger_and_enabled(0, found_trigger_value);
+	return is_mode_trigger_and_enabled(for_mode, 0, found_trigger_value);
 }
 
 /* Returns 1 if the mode is 'trigger', or 'default', where the default mode for
  * a feature is to trigger. Does not check whether a trigger is present. */
-int xdebug_lib_start_if_mode_is_trigger(void)
+int xdebug_lib_start_if_mode_is_trigger(int for_mode)
 {
-	return is_mode_trigger_and_enabled(1, NULL);
+	return is_mode_trigger_and_enabled(for_mode, 1, NULL);
 }
 
 
@@ -588,6 +706,18 @@ void xdebug_llist_string_dtor(void *dummy, void *elem)
 	}
 }
 
+char* xdebug_wrap_location_around_function_name(const char *prefix, zend_op_array *opa, char *fname)
+{
+	return xdebug_sprintf(
+		"%s{%s:%s:%d-%d}",
+		fname,
+		prefix,
+		opa->filename->val,
+		opa->line_start,
+		opa->line_end
+	);
+}
+
 char* xdebug_wrap_closure_location_around_function_name(zend_op_array *opa, char *fname)
 {
 	xdebug_str tmp = XDEBUG_STR_INITIALIZER;
@@ -611,3 +741,25 @@ char* xdebug_wrap_closure_location_around_function_name(zend_op_array *opa, char
 
 	return tmp.d;
 }
+
+static void xdebug_declared_var_dtor(void *dummy, void *elem)
+{
+	xdebug_str *s = (xdebug_str*) elem;
+
+	xdebug_str_free(s);
+}
+
+void xdebug_lib_register_compiled_variables(function_stack_entry *fse, zend_op_array *op_array)
+{
+	unsigned int i = 0;
+
+	if (!fse->declared_vars) {
+		fse->declared_vars = xdebug_llist_alloc(xdebug_declared_var_dtor);
+	}
+
+	/* gather used variables from compiled vars information */
+	while (i < (unsigned int) op_array->last_var) {
+		xdebug_llist_insert_next(fse->declared_vars, XDEBUG_LLIST_TAIL(fse->declared_vars), xdebug_str_create(STR_NAME_VAL(op_array->vars[i]), STR_NAME_LEN(op_array->vars[i])));
+		i++;
+	}
+}
